from abc import (
    ABC,
)
from typing import (
    Final,
    Iterable,
)

from .abc import (
    EnrouteDecorator,
)
from .kinds import (
    EnrouteDecoratorKind,
)


class RestEnrouteDecorator(EnrouteDecorator, ABC):
    """Rest Enroute class"""

    def __init__(self, url: str, method: str):
        self.url = url
        self.method = method

    def __iter__(self) -> Iterable:
        yield from (
            self.url,
            self.method,
        )


class RestCommandEnrouteDecorator(RestEnrouteDecorator):
    """Rest Command Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Command


class RestQueryEnrouteDecorator(RestEnrouteDecorator):
    """Rest Query Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Query



from __future__ import (
    annotations,
)

import logging
from functools import (
    cached_property,
)
from inspect import (
    isawaitable,
)
from typing import (
    Awaitable,
    Callable,
    Optional,
    Union,
)

from aiohttp import (
    web,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
)

from ..decorators import (
    EnrouteBuilder,
)
from ..requests import (
    REQUEST_USER_CONTEXT_VAR,
    Response,
    ResponseException,
)
from .requests import (
    RestRequest,
    RestResponse,
)

logger = logging.getLogger(__name__)


class RestHandler(MinosSetup):
    """Rest Handler class."""

    def __init__(self, host: str, port: int, endpoints: dict[(str, str), Callable], **kwargs):
        super().__init__(**kwargs)
        self._host = host
        self._port = port
        self._endpoints = endpoints

    @property
    def endpoints(self) -> dict[(str, str), Callable]:
        """Endpoints getter.

        :return: A dictionary value.
        """
        return self._endpoints

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> RestHandler:
        host = config.rest.host
        port = config.rest.port
        endpoints = cls._endpoints_from_config(config)

        return cls(host=host, port=port, endpoints=endpoints, **kwargs)

    @staticmethod
    def _endpoints_from_config(config: MinosConfig, **kwargs) -> dict[(str, str), Callable]:
        builder = EnrouteBuilder(*config.services, middleware=config.middleware)
        decorators = builder.get_rest_command_query(config=config, **kwargs)
        endpoints = {(decorator.url, decorator.method): fn for decorator, fn in decorators.items()}
        return endpoints

    @property
    def host(self) -> str:
        """Get the rest host.

        :return: A ``str`` object.
        """
        return self._host

    @property
    def port(self) -> int:
        """Get the rest port.

        :return: An integer value.
        """
        return self._port

    def get_app(self) -> web.Application:
        """Return rest application instance.

        :return: A `web.Application` instance.
        """
        return self._app

    @cached_property
    def _app(self) -> web.Application:
        app = web.Application()
        self._mount_routes(app)
        return app

    def _mount_routes(self, app: web.Application):
        """Load routes from config file."""
        for (url, method), fn in self._endpoints.items():
            self._mount_one_route(method, url, fn, app)

        # Load default routes
        self._mount_system_health(app)

    def _mount_one_route(self, method: str, url: str, action: Callable, app: web.Application) -> None:
        handler = self.get_callback(action)
        app.router.add_route(method, url, handler)

    @staticmethod
    def get_callback(
        fn: Callable[[RestRequest], Union[Optional[RestResponse], Awaitable[Optional[RestResponse]]]]
    ) -> Callable[[web.Request], Awaitable[web.Response]]:
        """Get the handler function to be used by the ``aiohttp`` Controller.

        :param fn: The action function.
        :return: A wrapper function around the given one that is compatible with the ``aiohttp`` Controller.
        """

        async def _fn(request: web.Request) -> web.Response:
            logger.info(f"Dispatching '{request!s}' from '{request.remote!s}'...")

            request = RestRequest(request)
            token = REQUEST_USER_CONTEXT_VAR.set(request.user)

            try:
                response = fn(request)
                if isawaitable(response):
                    response = await response
                if isinstance(response, Response):
                    response = await response.raw_content()
                if response is None:
                    return web.json_response()
                return web.json_response(response)
            except ResponseException as exc:
                logger.warning(f"Raised an application exception: {exc!s}")
                raise web.HTTPBadRequest(text=str(exc))
            except Exception as exc:
                logger.exception(f"Raised a system exception: {exc!r}")
                raise web.HTTPInternalServerError()
            finally:
                REQUEST_USER_CONTEXT_VAR.reset(token)

        return _fn

    def _mount_system_health(self, app: web.Application):
        """Mount System Health Route."""
        app.router.add_get("/system/health", self._system_health_handler)

    @staticmethod
    async def _system_health_handler(request: web.Request) -> web.Response:
        """System Health Route Handler.
        :return: A `web.json_response` response.
        """
        logger.info(f"Dispatching '{request!s}' from '{request.remote!s}'...")
        return web.json_response({"host": request.host})



from psycopg2.sql import (
    SQL,
)

from minos.common import (
    PostgreSqlMinosDatabase,
)


class BrokerHandlerSetup(PostgreSqlMinosDatabase):
    """Minos Broker Setup Class"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def _setup(self) -> None:
        await self._create_event_queue_table()

    async def _create_event_queue_table(self) -> None:
        _CREATE_TABLE_QUERY = SQL(
            "CREATE TABLE IF NOT EXISTS consumer_queue ("
            '"id" BIGSERIAL NOT NULL PRIMARY KEY, '
            '"topic" VARCHAR(255) NOT NULL, '
            '"partition" INTEGER,'
            '"data" BYTEA NOT NULL, '
            '"retry" INTEGER NOT NULL DEFAULT 0,'
            '"processing" BOOL NOT NULL DEFAULT FALSE, '
            '"created_at" TIMESTAMPTZ NOT NULL DEFAULT NOW(), '
            '"updated_at" TIMESTAMPTZ NOT NULL DEFAULT NOW())'
        )
        await self.submit_query(_CREATE_TABLE_QUERY, lock=hash("consumer_queue"))



from __future__ import (
    annotations,
)

from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from ...requests import (
    Request,
    Response,
    ResponseException,
)
from ..messages import (
    BrokerMessage,
)


class BrokerRequest(Request):
    """Handler Request class."""

    __slots__ = "raw"

    def __init__(self, raw: BrokerMessage):
        self.raw = raw

    def __eq__(self, other: BrokerRequest) -> bool:
        return type(self) == type(other) and self.raw == other.raw

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.raw!r})"

    @property
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        return getattr(self.raw, "user", None)

    async def content(self, **kwargs) -> Any:
        """Request content.

        :param kwargs: Additional named arguments.
        :return: The content.
        """
        data = self.raw.data
        return data


class BrokerResponse(Response):
    """Handler Response class."""


class BrokerResponseException(ResponseException):
    """Handler Response Exception class."""



from enum import (
    Enum,
    auto,
)


class EnrouteDecoratorKind(Enum):
    """Enroute Kind enumerate."""

    Command = auto()
    Query = auto()
    Event = auto()

    @property
    def pre_fn_name(self) -> str:
        """Get the pre execution function name.

        :return: A string value containing the function name.
        """
        mapping = {
            self.Command: "_pre_command_handle",
            self.Query: "_pre_query_handle",
            self.Event: "_pre_event_handle",
        }
        return mapping[self]

    @property
    def post_fn_name(self) -> str:
        """Get the post execution function name.

        :return: A string value containing the function name.
        """
        mapping = {
            self.Command: "_post_command_handle",
            self.Query: "_post_query_handle",
            self.Event: "_post_event_handle",
        }
        return mapping[self]



import logging
from typing import (
    Any,
)

from aiomisc import (
    Service,
)
from cached_property import (
    cached_property,
)
from dependency_injector.wiring import (
    Provide,
    inject,
)

from minos.common import (
    NotProvidedException,
)

from .consumers import (
    BrokerConsumer,
)
from .handlers import (
    BrokerHandler,
)

logger = logging.getLogger(__name__)


class BrokerConsumerService(Service):
    """Broker Consumer Service class."""

    @inject
    def __init__(self, consumer: BrokerConsumer = Provide["broker_consumer"], **kwargs):
        super().__init__(**kwargs)
        if consumer is None or isinstance(consumer, Provide):
            raise NotProvidedException(f"A {BrokerConsumer!r} object must be provided.")
        self.consumer = consumer

    async def start(self) -> None:
        """Start the service execution.

        :return: This method does not return anything.
        """
        await self.consumer.setup()

        try:
            self.start_event.set()
        except RuntimeError:
            logger.warning("Runtime is not properly setup.")

        await self.consumer.dispatch()

    async def stop(self, exception: Exception = None) -> Any:
        """Stop the service execution.

        :param exception: Optional exception that stopped the execution.
        :return: This method does not return anything.
        """
        await self.consumer.destroy()


class BrokerHandlerService(Service):
    """Broker Handler Service class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_kwargs = kwargs

    async def start(self) -> None:
        """Start the service execution.

        :return: This method does not return anything.
        """
        await self.handler.setup()

        try:
            self.start_event.set()
        except RuntimeError:
            logger.warning("Runtime is not properly setup.")

        await self.handler.dispatch_forever()

    async def stop(self, err: Exception = None) -> None:
        """Stop the service execution.

        :param err: Optional exception that stopped the execution.
        :return: This method does not return anything.
        """
        await self.handler.destroy()

    @cached_property
    def handler(self) -> BrokerHandler:
        """Get the service handler.

        :return: A ``Handler`` instance.
        """
        return BrokerHandler.from_config(**self._init_kwargs)



__author__ = """Clariteia Devs"""
__email__ = "devs@clariteia.com"
__version__ = "0.3.1"

from .brokers import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
    BrokerConsumer,
    BrokerConsumerService,
    BrokerHandler,
    BrokerHandlerEntry,
    BrokerHandlerService,
    BrokerHandlerSetup,
    BrokerMessage,
    BrokerMessageStatus,
    BrokerMessageStrategy,
    BrokerProducer,
    BrokerProducerService,
    BrokerPublisher,
    BrokerPublisherSetup,
    BrokerRequest,
    BrokerResponse,
    BrokerResponseException,
    DynamicBroker,
    DynamicBrokerPool,
)
from .decorators import (
    BrokerCommandEnrouteDecorator,
    BrokerEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
    EnrouteAnalyzer,
    EnrouteBuilder,
    EnrouteDecorator,
    EnrouteDecoratorKind,
    PeriodicEnrouteDecorator,
    PeriodicEventEnrouteDecorator,
    RestCommandEnrouteDecorator,
    RestEnrouteDecorator,
    RestQueryEnrouteDecorator,
    enroute,
)
from .discovery import (
    DiscoveryClient,
    DiscoveryConnector,
    KongDiscoveryClient,
    MinosDiscoveryClient,
)
from .exceptions import (
    MinosActionNotFoundException,
    MinosDiscoveryConnectorException,
    MinosHandlerException,
    MinosHandlerNotFoundEnoughEntriesException,
    MinosInvalidDiscoveryClient,
    MinosMultipleEnrouteDecoratorKindsException,
    MinosNetworkException,
    MinosRedefinedEnrouteDecoratorException,
)
from .requests import (
    REQUEST_USER_CONTEXT_VAR,
    Request,
    Response,
    ResponseException,
    WrappedRequest,
)
from .rest import (
    RestHandler,
    RestRequest,
    RestResponse,
    RestResponseException,
    RestService,
)
from .scheduling import (
    PeriodicTask,
    PeriodicTaskScheduler,
    PeriodicTaskSchedulerService,
    ScheduledRequest,
    ScheduledRequestContent,
    ScheduledResponseException,
)
from .utils import (
    consume_queue,
    get_host_ip,
    get_host_name,
    get_ip,
)



from __future__ import (
    annotations,
)

import logging
from inspect import (
    isclass,
)
from itertools import (
    chain,
)
from operator import (
    itemgetter,
)
from typing import (
    Any,
    Type,
)

from minos.common import (
    MinosConfig,
    MinosImportException,
    MinosSetup,
    import_module,
)

from ..decorators import (
    EnrouteAnalyzer,
)
from ..exceptions import (
    MinosInvalidDiscoveryClient,
)
from ..utils import (
    get_host_ip,
)
from .clients import (
    DiscoveryClient,
)

logger = logging.getLogger(__name__)


class DiscoveryConnector(MinosSetup):
    """Discovery Connector class."""

    def __init__(self, client, name: str, host: str, port: int, endpoints: list[dict[str, Any]], *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.client = client

        self.name = name
        self.host = host
        self.port = port
        self.endpoints = endpoints

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> DiscoveryConnector:
        client_cls = cls._client_cls_from_config(config)
        client = client_cls(host=config.discovery.host, port=config.discovery.port)
        port = config.rest.port
        name = config.service.name
        host = get_host_ip()
        endpoints = cls._endpoints_from_config(config)

        return cls(client, name, host, port, endpoints, *args, **kwargs)

    @staticmethod
    def _client_cls_from_config(config: MinosConfig) -> Type[DiscoveryClient]:
        try:
            # noinspection PyTypeChecker
            client_cls: type = import_module(config.discovery.client)
        except MinosImportException:
            raise MinosInvalidDiscoveryClient(f"{config.discovery.client} could not be imported.")

        if not isclass(client_cls) or not issubclass(client_cls, DiscoveryClient):
            raise MinosInvalidDiscoveryClient(f"{config.discovery.client} not supported.")
        return client_cls

    @staticmethod
    def _endpoints_from_config(config: MinosConfig) -> list[dict[str, Any]]:
        endpoints = list()
        for name in config.services:
            decorators = EnrouteAnalyzer(name, config).get_rest_command_query()
            endpoints += [
                {"url": decorator.url, "method": decorator.method} for decorator in set(chain(*decorators.values()))
            ]

        endpoints.sort(key=itemgetter("url", "method"))
        return endpoints

    async def _setup(self) -> None:
        await self.subscribe()

    async def subscribe(self) -> None:
        """Send a subscribe operation to the discovery.

        :return: This method does not return anything.
        """
        logger.info("Performing discovery subscription...")
        await self.client.subscribe(self.host, self.port, self.name, self.endpoints)

    async def _destroy(self) -> None:
        await self.unsubscribe()

    async def unsubscribe(self) -> None:
        """Send an unsubscribe operation to the discovery.

        :return: This method does not return anything.
        """
        logger.info("Performing discovery unsubscription...")
        await self.client.unsubscribe(self.name)



from __future__ import (
    annotations,
)

import asyncio
import logging
from contextlib import (
    suppress,
)
from datetime import (
    datetime,
)
from inspect import (
    isawaitable,
)
from typing import (
    Awaitable,
    Callable,
    NoReturn,
    Optional,
    Union,
)

from crontab import (
    CronTab,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
    current_datetime,
)

from ..decorators import (
    EnrouteBuilder,
)
from ..requests import (
    ResponseException,
)
from .requests import (
    ScheduledRequest,
)

logger = logging.getLogger(__name__)


class PeriodicTaskScheduler(MinosSetup):
    """Periodic Task Scheduler class."""

    def __init__(self, tasks: set[PeriodicTask], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tasks = tasks

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> PeriodicTaskScheduler:
        tasks = cls._tasks_from_config(config, **kwargs)
        return cls(tasks, **kwargs)

    @staticmethod
    def _tasks_from_config(config: MinosConfig, **kwargs) -> set[PeriodicTask]:
        builder = EnrouteBuilder(*config.services, middleware=config.middleware)
        decorators = builder.get_periodic_event(config=config, **kwargs)
        tasks = {PeriodicTask(decorator.crontab, fn) for decorator, fn in decorators.items()}
        return tasks

    @property
    def tasks(self) -> set[PeriodicTask]:
        """Get the set of periodic tasks.

        :return: A ``set`` of ``PeriodicTask`` instances.
        """
        return self._tasks

    async def start(self) -> None:
        """Start the execution of periodic tasks.

        :return: This method does not return anything.
        """

        await asyncio.gather(*(task.start() for task in self._tasks))

    async def stop(self, timeout: Optional[float] = None) -> None:
        """Stop the execution of periodic tasks.

        :param timeout: An optional timeout expressed in seconds.
        :return: This method does not return anything.
        """
        await asyncio.gather(*(task.stop(timeout=timeout) for task in self._tasks))


class PeriodicTask:
    """Periodic Task class."""

    _task: Optional[asyncio.Task]

    def __init__(self, crontab: Union[str, CronTab], fn: Callable[[ScheduledRequest], Awaitable[None]]):
        if isinstance(crontab, str):
            crontab = CronTab(crontab)

        self._crontab = crontab
        self._fn = fn
        self._task = None
        self._running = False

    @property
    def crontab(self) -> CronTab:
        """Get the crontab of the periodic task.

        :return: A ``CronTab`` instance.
        """
        return self._crontab

    @property
    def fn(self) -> Callable[[ScheduledRequest], Awaitable[None]]:
        """Get the function to be called periodically.

        :return: A function returning an awaitable.
        """
        return self._fn

    @property
    def started(self) -> bool:
        """Check if the periodic task has been started.

        :return: ``True`` if started or ``False`` otherwise.
        """

        return self._task is not None

    @property
    def task(self) -> asyncio.Task:
        """Get the asyncio task.

        :return: An ``asyncio.Task`` instance.
        """
        return self._task

    async def start(self) -> None:
        """Start the periodic task.

        :return: This method does not return anything.
        """
        logger.info("Starting periodic task...")
        self._task = asyncio.create_task(self.run_forever())

    async def stop(self, timeout: Optional[float] = None) -> None:
        """Stop the periodic task.

        :param timeout: An optional timeout expressed in seconds.
        :return: This method does not return anything.
        """
        if self._task is not None:
            logger.info("Stopping periodic task...")
            self._task.cancel()
            with suppress(asyncio.TimeoutError, asyncio.CancelledError):
                await asyncio.wait_for(self._task, timeout)
            self._task = None

    async def run_forever(self) -> NoReturn:
        """Run the periodic function forever. This method is equivalent to start, but it keeps waiting until infinite.

        :return: This method never returns.
        """
        now = current_datetime()
        await asyncio.sleep(self._crontab.next(now))

        while True:
            now = current_datetime()
            await asyncio.gather(asyncio.sleep(self._crontab.next(now)), self.run_once(now))

    @property
    def running(self) -> bool:
        """Check if the periodic function is running.

        :return: ``True`` if it's running or ``False`` otherwise.
        """
        return self._running

    async def run_once(self, now: Optional[datetime] = None) -> None:
        """Run the periodic function one time.

        :param now: An optional datetime expressing the current datetime.
        :return: This method does not return anything.
        """
        if now is None:
            now = current_datetime()

        request = ScheduledRequest(now)
        logger.debug("Running periodic task...")
        try:
            self._running = True
            with suppress(asyncio.CancelledError):
                response = self._fn(request)
                if isawaitable(response):
                    await response
        except ResponseException as exc:
            logger.warning(f"Raised an application exception: {exc!s}")
        except Exception as exc:
            logger.exception(f"Raised a system exception: {exc!r}")
        finally:
            self._running = False



from __future__ import (
    annotations,
)

import logging
from datetime import (
    datetime,
)
from functools import (
    total_ordering,
)
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    Type,
    TypeVar,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    MinosException,
    Model,
    current_datetime,
)

logger = logging.getLogger(__name__)
T = TypeVar("T")


@total_ordering
class BrokerHandlerEntry(Generic[T]):
    """Handler Entry class."""

    def __init__(
        self,
        id: int,
        topic: str,
        partition: int,
        data_bytes: bytes,
        retry: int = 0,
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
        data_cls: Type[Model] = Model,
        callback_lookup: Optional[Callable] = None,
        exception: Optional[Exception] = None,
    ):
        if created_at is None or updated_at is None:
            now = current_datetime()
            if created_at is None:
                created_at = now
            if updated_at is None:
                updated_at = now

        self.id = id
        self.topic = topic
        self.partition = partition
        self.data_bytes = data_bytes
        self.data_cls = data_cls
        self.retry = retry
        self.created_at = created_at
        self.updated_at = updated_at
        self.callback_lookup = callback_lookup
        self.exception = exception

    @property
    def success(self) -> bool:
        """Check if the entry is in success state or not

        :return: A boolean value.
        """
        return self.exception is None

    @cached_property
    def callback(self) -> Optional[Callable]:
        """Get the callback if lookup is provided.

        :return: A Callable object or None.
        """
        if self.callback_lookup is None:
            return None
        return self.callback_lookup(self.topic)

    @cached_property
    def data(self) -> T:
        """Get the data.

        :return: A ``Model`` inherited instance.
        """
        return self.data_cls.from_avro_bytes(self.data_bytes)

    def __lt__(self, other: Any) -> bool:
        # noinspection PyBroadException
        try:
            return isinstance(other, type(self)) and self.data.data < other.data.data
        except Exception:
            return False

    def __eq__(self, other):
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __iter__(self):
        yield from (
            self.id,
            self.topic,
            self.partition,
            self.data_bytes,
            self.data_cls,
            self.retry,
            self.created_at,
            self.updated_at,
            self.callback_lookup,
            self.exception,
        )

    def __repr__(self):
        try:
            return f"{type(self).__name__}(id={self.id!r}, data={self.data!r})"
        except MinosException:
            return f"{type(self).__name__}(id={self.id!r})"



from .dynamic import (
    DynamicBroker,
    DynamicBrokerPool,
)
from .handlers import (
    BrokerConsumer,
    BrokerConsumerService,
    BrokerHandler,
    BrokerHandlerEntry,
    BrokerHandlerService,
    BrokerHandlerSetup,
    BrokerRequest,
    BrokerResponse,
    BrokerResponseException,
)
from .messages import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
    BrokerMessage,
    BrokerMessageStatus,
    BrokerMessageStrategy,
)
from .publishers import (
    BrokerProducer,
    BrokerProducerService,
    BrokerPublisher,
    BrokerPublisherSetup,
)



import re
import socket
from asyncio import (
    QueueEmpty,
)


def get_host_ip() -> str:
    """Get the host ip.

    :return: A string value.
    """
    name = get_host_name()
    return get_ip(name)


def get_host_name() -> str:
    """Get the host name.

    :return: A string value.
    """
    return re.sub(r"\.(?:local|lan)", "", socket.gethostname())


def get_ip(name: str) -> str:
    """Get the ip address.

    :param name: The name to be converted to an ip.
    :return: A string value.
    """
    return socket.gethostbyname(name)


async def consume_queue(queue, max_count: int) -> None:
    """Consume ``count`` at least ``1`` and at most ``max_count`` elements from the given queue.

    :param queue: The queue to be consumed.
    :param max_count: The max count of elements to be consumed.
    :return: This function does not return anything.
    """
    await queue.get()

    c = 1
    while c < max_count:
        c += 1
        try:
            queue.get_nowait()
        except QueueEmpty:
            break



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    TimeoutError,
    wait_for,
)
from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from aiopg import (
    Cursor,
)
from cached_property import (
    cached_property,
)
from dependency_injector.wiring import (
    Provide,
    inject,
)
from psycopg2.sql import (
    SQL,
    Identifier,
)

from minos.common import (
    MinosConfig,
    NotProvidedException,
)

from ...exceptions import (
    MinosHandlerNotFoundEnoughEntriesException,
)
from ...utils import (
    consume_queue,
)
from ..handlers import (
    BrokerHandlerEntry,
    BrokerHandlerSetup,
)
from ..publishers import (
    BrokerPublisher,
)

logger = logging.getLogger(__name__)


class DynamicBroker(BrokerHandlerSetup):
    """Dynamic Broker class."""

    def __init__(self, topic: str, publisher: BrokerPublisher, **kwargs):
        super().__init__(**kwargs)

        self.topic = topic
        self.publisher = publisher

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> DynamicBroker:
        kwargs["publisher"] = cls._get_publisher(**kwargs)
        # noinspection PyProtectedMember
        return cls(**config.broker.queue._asdict(), **kwargs)

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_publisher(
        publisher: Optional[BrokerPublisher] = None,
        broker_publisher: BrokerPublisher = Provide["broker_publisher"],
        **kwargs,
    ) -> BrokerPublisher:
        if publisher is None:
            publisher = broker_publisher
        if publisher is None or isinstance(publisher, Provide):
            raise NotProvidedException(f"A {BrokerPublisher!r} object must be provided.")
        return publisher

    async def _setup(self) -> None:
        await super()._setup()

    async def _destroy(self) -> None:
        await super()._destroy()

    # noinspection PyUnusedLocal
    async def send(self, *args, reply_topic: None = None, **kwargs) -> UUID:
        """Send a ``BrokerMessage``.

        :param args: Additional positional arguments.
        :param reply_topic: This argument is ignored if ignored in favor of ``self.topic``.
        :param kwargs: Additional named arguments.
        :return: The ``UUID`` identifier of the message.
        """
        return await self.publisher.send(*args, reply_topic=self.topic, **kwargs)

    async def get_one(self, *args, **kwargs) -> BrokerHandlerEntry:
        """Get one handler entry from the given topics.

        :param args: Additional positional parameters to be passed to get_many.
        :param kwargs: Additional named parameters to be passed to get_many.
        :return: A ``HandlerEntry`` instance.
        """
        return (await self.get_many(*args, **(kwargs | {"count": 1})))[0]

    async def get_many(self, count: int, timeout: float = 60, **kwargs) -> list[BrokerHandlerEntry]:
        """Get multiple handler entries from the given topics.

        :param timeout: Maximum time in seconds to wait for messages.
        :param count: Number of entries to be collected.
        :return: A list of ``HandlerEntry`` instances.
        """
        try:
            entries = await wait_for(self._get_many(count, **kwargs), timeout=timeout)
        except TimeoutError:
            raise MinosHandlerNotFoundEnoughEntriesException(
                f"Timeout exceeded while trying to fetch {count!r} entries from {self.topic!r}."
            )

        logger.info(f"Dispatching '{entries if count > 1 else entries[0]!s}'...")

        return entries

    async def _get_many(self, count: int, max_wait: Optional[float] = 10.0) -> list[BrokerHandlerEntry]:
        result = list()
        async with self.cursor() as cursor:

            await cursor.execute(self._queries["listen"])
            try:
                while len(result) < count:
                    await self._wait_for_entries(cursor, count - len(result), max_wait)
                    result += await self._get_entries(cursor, count - len(result))
            finally:
                await cursor.execute(self._queries["unlisten"])

        return result

    async def _wait_for_entries(self, cursor: Cursor, count: int, max_wait: Optional[float]) -> None:
        if await self._get_count(cursor):
            return

        while True:
            try:
                return await wait_for(consume_queue(cursor.connection.notifies, count), max_wait)
            except TimeoutError:
                if await self._get_count(cursor):
                    return

    async def _get_count(self, cursor) -> int:
        await cursor.execute(self._queries["count_not_processed"], (self.topic,))
        count = (await cursor.fetchone())[0]
        return count

    async def _get_entries(self, cursor: Cursor, count: int) -> list[BrokerHandlerEntry]:
        entries = list()
        async with cursor.begin():
            await cursor.execute(self._queries["select_not_processed"], (self.topic, count))
            for entry in self._build_entries(await cursor.fetchall()):
                await cursor.execute(self._queries["delete_processed"], (entry.id,))
                entries.append(entry)
        return entries

    @cached_property
    def _queries(self) -> dict[str, str]:
        # noinspection PyTypeChecker
        return {
            "listen": _LISTEN_QUERY.format(Identifier(self.topic)),
            "unlisten": _UNLISTEN_QUERY.format(Identifier(self.topic)),
            "count_not_processed": _COUNT_NOT_PROCESSED_QUERY,
            "select_not_processed": _SELECT_NOT_PROCESSED_ROWS_QUERY,
            "delete_processed": _DELETE_PROCESSED_QUERY,
        }

    @staticmethod
    def _build_entries(rows: list[tuple]) -> list[BrokerHandlerEntry]:
        return [BrokerHandlerEntry(*row) for row in rows]


_LISTEN_QUERY = SQL("LISTEN {}")

_UNLISTEN_QUERY = SQL("UNLISTEN {}")

# noinspection SqlDerivedTableAlias
_COUNT_NOT_PROCESSED_QUERY = SQL(
    "SELECT COUNT(*) FROM (SELECT id FROM consumer_queue WHERE topic = %s FOR UPDATE SKIP LOCKED) s"
)

_SELECT_NOT_PROCESSED_ROWS_QUERY = SQL(
    "SELECT id, topic, partition, data, retry, created_at, updated_at "
    "FROM consumer_queue "
    "WHERE topic = %s "
    "ORDER BY created_at "
    "LIMIT %s "
    "FOR UPDATE SKIP LOCKED"
)
_DELETE_PROCESSED_QUERY = SQL("DELETE FROM consumer_queue WHERE id = %s")



from abc import (
    ABC,
)
from typing import (
    Final,
    Iterable,
)

from .abc import (
    EnrouteDecorator,
)
from .kinds import (
    EnrouteDecoratorKind,
)


class BrokerEnrouteDecorator(EnrouteDecorator, ABC):
    """Broker Enroute class"""

    def __init__(self, topic: str):
        self.topic = topic

    def __iter__(self) -> Iterable:
        yield from (self.topic,)


class BrokerCommandEnrouteDecorator(BrokerEnrouteDecorator):
    """Broker Command Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Command


class BrokerQueryEnrouteDecorator(BrokerEnrouteDecorator):
    """Broker Query Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Query


class BrokerEventEnrouteDecorator(BrokerEnrouteDecorator):
    """Broker Event Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Event



from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from psycopg2.sql import (
    SQL,
)

from minos.common import (
    MinosConfig,
)

from ..messages import (
    BrokerMessage,
    BrokerMessageStatus,
    BrokerMessageStrategy,
)
from .abc import (
    BrokerPublisherSetup,
)

logger = logging.getLogger(__name__)


class BrokerPublisher(BrokerPublisherSetup):
    """Broker Publisher class."""

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> BrokerPublisher:
        # noinspection PyProtectedMember
        return cls(*args, **config.broker.queue._asdict(), **kwargs)

    # noinspection PyMethodOverriding
    async def send(
        self,
        data: Any,
        topic: str,
        *,
        identifier: Optional[UUID] = None,
        reply_topic: Optional[str] = None,
        user: Optional[UUID] = None,
        status: BrokerMessageStatus = BrokerMessageStatus.SUCCESS,
        strategy: BrokerMessageStrategy = BrokerMessageStrategy.UNICAST,
        headers: Optional[dict[str, str]] = None,
        **kwargs,
    ) -> UUID:
        """Send a ``BrokerMessage``.

        :param data: The data to be send.
        :param topic: Topic in which the message will be published.
        :param identifier: The identifier of the message.
        :param reply_topic: An optional topic name to wait for a response.
        :param user: The user identifier that send the message.
        :param status: The status code of the message.
        :param strategy: The publishing strategy.
        :param headers: A mapping of string values identified by a string key.
        :param kwargs: Additional named arguments.
        :return: The ``UUID`` identifier of the message.
        """

        message = BrokerMessage(
            topic=topic,
            data=data,
            identifier=identifier,
            status=status,
            reply_topic=reply_topic,
            user=user,
            strategy=strategy,
            headers=headers,
        )
        logger.info(f"Publishing '{message!s}'...")
        await self.enqueue(message.topic, message.strategy, message.avro_bytes)
        return message.identifier

    async def enqueue(self, topic: str, strategy: BrokerMessageStrategy, raw: bytes) -> int:
        """Send a sequence of bytes to the given topic.

        :param topic: Topic in which the bytes will be send.
        :param strategy: The publishing strategy.
        :param raw: Bytes sequence to be send.
        :return: The identifier of the message in the queue.
        """
        params = (topic, raw, strategy)
        raw = await self.submit_query_and_fetchone(_INSERT_ENTRY_QUERY, params)
        await self.submit_query(_NOTIFY_QUERY)
        return raw[0]


_INSERT_ENTRY_QUERY = SQL("INSERT INTO producer_queue (topic, data, strategy) VALUES (%s, %s, %s) RETURNING id")

_NOTIFY_QUERY = SQL("NOTIFY producer_queue")



import logging
from abc import (
    ABC,
    abstractmethod,
)
from asyncio import (
    sleep,
)

import aiohttp

from ...exceptions import (
    MinosDiscoveryConnectorException,
)

logger = logging.getLogger(__name__)


class DiscoveryClient(ABC):
    """Minos Discovery Client class."""

    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port

    @property
    def route(self) -> str:
        """Get the full http route to the repository.

        :return: An ``str`` value.
        """
        # noinspection HttpUrlsUsage
        return f"http://{self.host}:{self.port}"

    @abstractmethod
    async def subscribe(
        self,
        host: str,
        port: int,
        name: str,
        endpoints: list[dict[str, str]],
        retry_tries: int = 3,
        retry_delay: float = 5,
    ) -> None:
        """ Perform a subscription query.

        :param host: The ip of the microservice to be subscribed.
        :param port: The port of the microservice to be subscribed.
        :param name: The name of the microservice to be subscribed.
        :param endpoints: List of endpoints exposed by the microservice.
        :param retry_tries: Number of attempts before raising a failure exception.
        :param retry_delay: Seconds to wait between attempts.
        :return: This method does not return anything.
        """

    async def _rest_subscribe(
        self,
        endpoint: str,
        service_metadata: dict[str, str],
        host: str,
        port: int,
        name: str,
        endpoints: list[dict[str, str]],
        retry_tries: int = 3,
        retry_delay: float = 5,
    ) -> None:
        logger.debug(f"Subscribing into {endpoint!r}...")
        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(endpoint, json=service_metadata) as response:
                    success = response.ok
        except Exception as exc:
            logger.warning(f"An exception was raised while trying to subscribe: {exc!r}")
            success = False

        if not success:
            if retry_tries > 1:
                await sleep(retry_delay)
                return await self.subscribe(host, port, name, endpoints, retry_tries - 1, retry_delay)
            else:
                raise MinosDiscoveryConnectorException("There was a problem while trying to subscribe.")

    @abstractmethod
    async def unsubscribe(self, name: str, retry_tries: int = 3, retry_delay: float = 5) -> None:
        """ Perform an unsubscribe query.

        :param name: The name of the microservice to be unsubscribed.
        :param retry_tries: Number of attempts before raising a failure exception.
        :param retry_delay: Seconds to wait between attempts.
        :return: This method does not return anything.
        """

    async def _rest_unsubscribe(self, endpoint: str, name: str, retry_tries: int = 3, retry_delay: float = 5) -> None:
        logger.debug(f"Unsubscribing into {endpoint!r}...")

        try:
            async with aiohttp.ClientSession() as session:
                async with session.delete(endpoint) as response:
                    success = response.ok
        except Exception as exc:
            logger.warning(f"An exception was raised while trying to unsubscribe: {exc!r}")
            success = False

        if not success:
            if retry_tries > 1:
                await sleep(retry_delay)
                return await self.unsubscribe(name, retry_tries - 1, retry_delay)
            else:
                raise MinosDiscoveryConnectorException("There was a problem while trying to unsubscribe.")



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    TimeoutError,
    gather,
    wait_for,
)
from typing import (
    NoReturn,
    Optional,
)

from aiokafka import (
    AIOKafkaProducer,
)
from aiopg import (
    Cursor,
)
from cached_property import (
    cached_property,
)
from dependency_injector.wiring import (
    Provide,
    inject,
)
from psycopg2.sql import (
    SQL,
)

from minos.common import (
    MinosConfig,
    NotProvidedException,
)

from ...utils import (
    consume_queue,
)
from ..handlers import (
    BrokerConsumer,
)
from ..messages import (
    BrokerMessageStrategy,
)
from .abc import (
    BrokerPublisherSetup,
)

logger = logging.getLogger(__name__)


class BrokerProducer(BrokerPublisherSetup):
    """Broker Producer class."""

    def __init__(
        self,
        *args,
        broker_host: str,
        broker_port: int,
        retry: int,
        records: int,
        client: Optional[AIOKafkaProducer] = None,
        consumer: BrokerConsumer = Provide["broker_consumer"],
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.retry = retry
        self.records = records
        self._client = client
        self.consumer = consumer

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> BrokerProducer:
        kwargs["broker_host"] = config.broker.host
        kwargs["broker_port"] = config.broker.port
        kwargs["consumer"] = cls._get_consumer(**kwargs)
        # noinspection PyProtectedMember
        return cls(**config.broker.queue._asdict(), **kwargs)

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_consumer(
        consumer: Optional[BrokerConsumer] = None,
        broker_consumer: BrokerConsumer = Provide["broker_consumer"],
        **kwargs,
    ) -> BrokerConsumer:
        if consumer is None:
            consumer = broker_consumer
        if consumer is None or isinstance(consumer, Provide):
            raise NotProvidedException(f"A {BrokerConsumer!r} object must be provided.")
        return consumer

    async def _setup(self) -> None:
        await super()._setup()
        await self.client.start()

    async def _destroy(self) -> None:
        await self.client.stop()
        await super()._destroy()

    async def dispatch_forever(self, max_wait: Optional[float] = 60.0) -> NoReturn:
        """Dispatch the items in the publishing queue forever.

        :param max_wait: Maximum seconds to wait for notifications. If ``None`` the wait is performed until infinity.
        :return: This method does not return anything.
        """
        async with self.cursor() as cursor:
            await cursor.execute(self._queries["listen"])
            try:
                while True:
                    await self._wait_for_entries(cursor, max_wait)
                    await self.dispatch(cursor)
            finally:
                await cursor.execute(self._queries["unlisten"])

    async def _wait_for_entries(self, cursor: Cursor, max_wait: Optional[float]) -> None:
        if await self._get_count(cursor):
            return

        while True:
            try:
                return await wait_for(consume_queue(cursor.connection.notifies, self.records), max_wait)
            except TimeoutError:
                if await self._get_count(cursor):
                    return

    async def _get_count(self, cursor) -> int:
        await cursor.execute(self._queries["count_not_processed"], (self.retry,))
        count = (await cursor.fetchone())[0]
        return count

    async def dispatch(self, cursor: Optional[Cursor] = None) -> None:
        """Dispatch the items in the publishing queue.

        :return: This method does not return anything.
        """
        is_external_cursor = cursor is not None
        if not is_external_cursor:
            cursor = await self.cursor().__aenter__()

        async with cursor.begin():
            await cursor.execute(self._queries["select_not_processed"], (self.retry, self.records))

            rows = await cursor.fetchall()
            futures = (self.dispatch_one(row) for row in rows)
            result = zip(await gather(*futures), rows)

            for (published, row) in result:
                query_id = "delete_processed" if published else "update_not_processed"
                await cursor.execute(self._queries[query_id], (row[0],))

        if not is_external_cursor:
            await cursor.__aexit__(None, None, None)

    @cached_property
    def _queries(self) -> dict[str, str]:
        # noinspection PyTypeChecker
        return {
            "listen": _LISTEN_QUERY,
            "unlisten": _UNLISTEN_QUERY,
            "count_not_processed": _COUNT_NOT_PROCESSED_QUERY,
            "select_not_processed": _SELECT_NOT_PROCESSED_QUERY,
            "delete_processed": _DELETE_PROCESSED_QUERY,
            "update_not_processed": _UPDATE_NOT_PROCESSED_QUERY,
        }

    async def dispatch_one(self, row: tuple) -> bool:
        """Dispatch one row.

        :param row: A row containing the message information.
        :return: ``True`` if everything was fine or ``False`` otherwise.
        """
        topic, message, strategy = row[1], row[2], row[3]

        # noinspection PyBroadException
        try:
            if strategy == BrokerMessageStrategy.UNICAST and topic in self.consumer.topics:
                await self.consumer.enqueue(topic, -1, message)
                return True

        except Exception as exc:
            logger.warning(f"There was a problem while trying to use the consumer: {exc!r}")

        return await self.publish(topic, message)

    async def publish(self, topic: str, message: bytes) -> bool:
        """Publish a new item in the broker (kafka).

        :param topic: The topic in which the message will be published.
        :param message: The message to be published.
        :return: A boolean flag, ``True`` when the message is properly published or ``False`` otherwise.
        """
        logger.debug(f"Producing message with {topic!s} topic...")

        # noinspection PyBroadException
        try:
            await self.client.send_and_wait(topic, message)
            return True
        except Exception:
            return False

    @property
    def client(self) -> AIOKafkaProducer:
        """Get the client instance.

        :return: An ``AIOKafkaProducer`` instance.
        """
        if self._client is None:  # pragma: no cover
            self._client = AIOKafkaProducer(bootstrap_servers=f"{self.broker_host}:{self.broker_port}")
        return self._client


# noinspection SqlDerivedTableAlias
_COUNT_NOT_PROCESSED_QUERY = SQL(
    "SELECT COUNT(*) FROM (SELECT id FROM producer_queue WHERE retry < %s FOR UPDATE SKIP LOCKED) s"
)

_SELECT_NOT_PROCESSED_QUERY = SQL(
    "SELECT * "
    "FROM producer_queue "
    "WHERE retry < %s "
    "ORDER BY created_at "
    "LIMIT %s "
    "FOR UPDATE "
    "SKIP LOCKED"
)

_DELETE_PROCESSED_QUERY = SQL("DELETE FROM producer_queue WHERE id = %s")

_UPDATE_NOT_PROCESSED_QUERY = SQL("UPDATE producer_queue SET retry = retry + 1, updated_at = NOW() WHERE id = %s")

_LISTEN_QUERY = SQL("LISTEN producer_queue")

_UNLISTEN_QUERY = SQL("UNLISTEN producer_queue")



from minos.common import (
    MinosException,
)


class MinosNetworkException(MinosException):
    """Base network exception."""


class MinosDiscoveryConnectorException(MinosNetworkException):
    """Exception to be raised when there is a failure while communicating with the discovery."""


class MinosInvalidDiscoveryClient(MinosNetworkException):
    """Exception raised when the configured Discovery Client does not implement de DiscoveryClient interface"""


class MinosHandlerException(MinosNetworkException):
    """Base handler exception."""


class MinosActionNotFoundException(MinosHandlerException):
    """Exception to be raised when an action cannot be found,"""


class MinosHandlerNotFoundEnoughEntriesException(MinosHandlerException):
    """Exception to be raised when not enough entries have been found by a handler."""


class MinosEnrouteDecoratorException(MinosNetworkException):
    """Base exception for enroute decorators."""


class MinosMultipleEnrouteDecoratorKindsException(MinosEnrouteDecoratorException):
    """Exception to be raised when multiple enroute decorator kinds are applied to the same function."""


class MinosRedefinedEnrouteDecoratorException(MinosEnrouteDecoratorException):
    """Exception to be raised when same enroute decorator is used by multiple handlers."""



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    CancelledError,
    PriorityQueue,
    Task,
    TimeoutError,
    create_task,
    gather,
    wait_for,
)
from inspect import (
    isawaitable,
)
from typing import (
    Any,
    Awaitable,
    Callable,
    KeysView,
    NoReturn,
    Optional,
    Union,
)

from aiopg import (
    Cursor,
)
from cached_property import (
    cached_property,
)
from dependency_injector.wiring import (
    Provide,
    inject,
)
from psycopg2.sql import (
    SQL,
    Identifier,
)

from minos.common import (
    MinosConfig,
    NotProvidedException,
)

from ...decorators import (
    EnrouteBuilder,
)
from ...exceptions import (
    MinosActionNotFoundException,
)
from ...requests import (
    REQUEST_USER_CONTEXT_VAR,
    Response,
    ResponseException,
)
from ...utils import (
    consume_queue,
)
from ..messages import (
    REQUEST_HEADERS_CONTEXT_VAR,
    BrokerMessage,
    BrokerMessageStatus,
)
from ..publishers import (
    BrokerPublisher,
)
from .abc import (
    BrokerHandlerSetup,
)
from .entries import (
    BrokerHandlerEntry,
)
from .requests import (
    BrokerRequest,
    BrokerResponse,
)

logger = logging.getLogger(__name__)


class BrokerHandler(BrokerHandlerSetup):
    """Broker Handler class."""

    __slots__ = "_handlers", "_records", "_retry", "_queue", "_consumers", "_consumer_concurrency"

    def __init__(
        self,
        records: int,
        handlers: dict[str, Optional[Callable]],
        retry: int,
        publisher: BrokerPublisher,
        consumer_concurrency: int = 15,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._handlers = handlers
        self._records = records
        self._retry = retry

        self._queue = PriorityQueue(maxsize=self._records)
        self._consumers: list[Task] = list()
        self._consumer_concurrency = consumer_concurrency

        self._publisher = publisher

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> BrokerHandler:
        kwargs["handlers"] = cls._get_handlers(config, **kwargs)
        kwargs["publisher"] = cls._get_publisher(**kwargs)
        # noinspection PyProtectedMember
        return cls(**config.broker.queue._asdict(), **kwargs)

    @staticmethod
    def _get_handlers(
        config: MinosConfig, handlers: dict[str, Optional[Callable]] = None, **kwargs
    ) -> dict[str, Callable[[BrokerRequest], Awaitable[Optional[BrokerResponse]]]]:
        if handlers is None:
            builder = EnrouteBuilder(*config.services, middleware=config.middleware)
            decorators = builder.get_broker_command_query_event(config=config, **kwargs)
            handlers = {decorator.topic: fn for decorator, fn in decorators.items()}
        return handlers

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_publisher(
        publisher: Optional[BrokerPublisher] = None,
        broker_publisher: BrokerPublisher = Provide["broker_publisher"],
        **kwargs,
    ) -> BrokerPublisher:
        if publisher is None:
            publisher = broker_publisher
        if publisher is None or isinstance(publisher, Provide):
            raise NotProvidedException(f"A {BrokerPublisher!r} object must be provided.")
        return publisher

    async def _setup(self) -> None:
        await super()._setup()
        await self._create_consumers()

    async def _destroy(self) -> None:
        await self._destroy_consumers()
        await super()._destroy()

    async def _create_consumers(self):
        while len(self._consumers) < self._consumer_concurrency:
            self._consumers.append(create_task(self._consume()))

    async def _destroy_consumers(self):
        for consumer in self._consumers:
            consumer.cancel()
        await gather(*self._consumers, return_exceptions=True)
        self._consumers = list()

        while not self._queue.empty():
            entry = self._queue.get_nowait()
            await self.submit_query(self._queries["update_not_processed"], (entry.id,))

    async def _consume(self) -> None:
        while True:
            await self._consume_one()

    async def _consume_one(self) -> None:
        entry = await self._queue.get()
        try:
            await self._dispatch_one(entry)
        finally:
            self._queue.task_done()

    @property
    def publisher(self) -> BrokerPublisher:
        """Get the publisher instance.

        :return: A ``BrokerPublisher`` instance.
        """
        return self._publisher

    @property
    def consumers(self) -> list[Task]:
        """Get the consumers.

        :return: A list of ``Task`` instances.
        """
        return self._consumers

    @property
    def handlers(self) -> dict[str, Optional[Callable]]:
        """Handlers getter.

        :return: A dictionary in which the keys are topics and the values are the handler.
        """
        return self._handlers

    @property
    def topics(self) -> KeysView[str]:
        """Get an iterable containing the topic names.

        :return: An ``Iterable`` of ``str``.
        """
        return self.handlers.keys()

    async def dispatch_forever(self, max_wait: Optional[float] = 60.0) -> NoReturn:
        """Dispatch the items in the consuming queue forever.

        :param max_wait: Maximum seconds to wait for notifications. If ``None`` the wait is performed until infinity.
        :return: This method does not return anything.
        """
        async with self.cursor() as cursor:
            await self._listen_entries(cursor)
            try:
                while True:
                    await self._wait_for_entries(cursor, max_wait)
                    await self.dispatch(cursor, background_mode=True)
            finally:
                await self._unlisten_entries(cursor)

    async def _listen_entries(self, cursor: Cursor):
        for topic in self.topics:
            # noinspection PyTypeChecker
            await cursor.execute(_LISTEN_QUERY.format(Identifier(topic)))

    async def _unlisten_entries(self, cursor: Cursor) -> None:
        for topic in self.topics:
            # noinspection PyTypeChecker
            await cursor.execute(_UNLISTEN_QUERY.format(Identifier(topic)))

    async def _wait_for_entries(self, cursor: Cursor, max_wait: Optional[float]) -> None:
        if await self._get_count(cursor):
            return

        while True:
            try:
                return await wait_for(consume_queue(cursor.connection.notifies, self._records), max_wait)
            except TimeoutError:
                if await self._get_count(cursor):
                    return

    async def _get_count(self, cursor) -> int:
        if not len(self.topics):
            return 0
        await cursor.execute(_COUNT_NOT_PROCESSED_QUERY, (self._retry, tuple(self.topics)))
        count = (await cursor.fetchone())[0]
        return count

    async def dispatch(self, cursor: Optional[Cursor] = None, background_mode: bool = False) -> None:
        """Dispatch a batch of ``HandlerEntry`` instances from the database's queue.

        :param cursor: The cursor to interact with the database. If ``None`` is provided a new one is acquired.
        :param background_mode: If ``True`` the entries dispatching waits until every entry is processed. Otherwise,
            the dispatching is performed on background.
        :return: This method does not return anything.
        """

        is_external_cursor = cursor is not None
        if not is_external_cursor:
            cursor = await self.cursor().__aenter__()

        async with cursor.begin():
            await cursor.execute(
                self._queries["select_not_processed"], (self._retry, tuple(self.topics), self._records)
            )
            result = await cursor.fetchall()

            if len(result):
                entries = self._build_entries(result)

                await cursor.execute(self._queries["mark_processing"], (tuple(e.id for e in entries),))

                for entry in entries:
                    await self._queue.put(entry)

        if not is_external_cursor:
            await cursor.__aexit__(None, None, None)

        if not background_mode:
            await self._queue.join()

    def _build_entries(self, rows: list[tuple]) -> list[BrokerHandlerEntry]:
        kwargs = {"callback_lookup": self.get_action}
        return [BrokerHandlerEntry(*row, **kwargs) for row in rows]

    async def _dispatch_one(self, entry: BrokerHandlerEntry) -> None:
        logger.debug(f"Dispatching '{entry!r}'...")
        try:
            await self.dispatch_one(entry)
        except (CancelledError, Exception) as exc:
            logger.warning(f"Raised an exception while dispatching {entry!r}: {exc!r}")
            entry.exception = exc
            if isinstance(exc, CancelledError):
                raise exc
        finally:
            query_id = "delete_processed" if entry.success else "update_not_processed"
            await self.submit_query(self._queries[query_id], (entry.id,))

    async def dispatch_one(self, entry: BrokerHandlerEntry) -> None:
        """Dispatch one row.

        :param entry: Entry to be dispatched.
        :return: This method does not return anything.
        """
        logger.info(f"Dispatching '{entry!s}'...")

        fn = self.get_callback(entry.callback)
        message = entry.data
        data, status, headers = await fn(message)

        if message.reply_topic is not None:
            await self.publisher.send(
                data,
                topic=message.reply_topic,
                identifier=message.identifier,
                status=status,
                user=message.user,
                headers=headers,
            )

    @staticmethod
    def get_callback(
        fn: Callable[[BrokerRequest], Union[Optional[BrokerRequest], Awaitable[Optional[BrokerRequest]]]]
    ) -> Callable[[BrokerMessage], Awaitable[tuple[Any, BrokerMessageStatus, dict[str, str]]]]:
        """Get the handler function to be used by the Broker Handler.

        :param fn: The action function.
        :return: A wrapper function around the given one that is compatible with the Broker Handler API.
        """

        async def _fn(raw: BrokerMessage) -> tuple[Any, BrokerMessageStatus, dict[str, str]]:
            request = BrokerRequest(raw)
            user_token = REQUEST_USER_CONTEXT_VAR.set(request.user)
            headers_token = REQUEST_HEADERS_CONTEXT_VAR.set(raw.headers)

            try:
                response = fn(request)
                if isawaitable(response):
                    response = await response
                if isinstance(response, Response):
                    response = await response.content()
                return response, BrokerMessageStatus.SUCCESS, REQUEST_HEADERS_CONTEXT_VAR.get()
            except ResponseException as exc:
                logger.warning(f"Raised an application exception: {exc!s}")
                return repr(exc), BrokerMessageStatus.ERROR, REQUEST_HEADERS_CONTEXT_VAR.get()
            except Exception as exc:
                logger.exception(f"Raised a system exception: {exc!r}")
                return repr(exc), BrokerMessageStatus.SYSTEM_ERROR, REQUEST_HEADERS_CONTEXT_VAR.get()
            finally:
                REQUEST_USER_CONTEXT_VAR.reset(user_token)
                REQUEST_HEADERS_CONTEXT_VAR.reset(headers_token)

        return _fn

    def get_action(self, topic: str) -> Optional[Callable]:
        """Get handling function to be called.

        Gets the instance of the class and method to call.

        Args:
            topic: Kafka topic. Example: "TicketAdded"

        Raises:
            MinosNetworkException: topic TicketAdded have no controller/action configured, please review th
                configuration file.
        """
        if topic not in self._handlers:
            raise MinosActionNotFoundException(
                f"topic {topic} have no controller/action configured, " f"please review th configuration file"
            )

        handler = self._handlers[topic]

        logger.debug(f"Loaded {handler!r} action!")
        return handler

    @cached_property
    def _queries(self) -> dict[str, str]:
        # noinspection PyTypeChecker
        return {
            "count_not_processed": _COUNT_NOT_PROCESSED_QUERY,
            "select_not_processed": _SELECT_NOT_PROCESSED_QUERY,
            "mark_processing": _MARK_PROCESSING_QUERY,
            "delete_processed": _DELETE_PROCESSED_QUERY,
            "update_not_processed": _UPDATE_NOT_PROCESSED_QUERY,
        }


# noinspection SqlDerivedTableAlias
_COUNT_NOT_PROCESSED_QUERY = SQL(
    "SELECT COUNT(*) "
    "FROM (SELECT id FROM consumer_queue WHERE NOT processing AND retry < %s AND topic IN %s FOR UPDATE SKIP LOCKED) s"
)

_SELECT_NOT_PROCESSED_QUERY = SQL(
    "SELECT id, topic, partition, data, retry, created_at, updated_at "
    "FROM consumer_queue "
    "WHERE NOT processing AND retry < %s AND topic IN %s "
    "ORDER BY created_at "
    "LIMIT %s "
    "FOR UPDATE SKIP LOCKED"
)

_MARK_PROCESSING_QUERY = SQL("UPDATE consumer_queue SET processing = TRUE WHERE id IN %s")

_DELETE_PROCESSED_QUERY = SQL("DELETE FROM consumer_queue WHERE id = %s")

_UPDATE_NOT_PROCESSED_QUERY = SQL(
    "UPDATE consumer_queue SET processing = FALSE, retry = retry + 1, updated_at = NOW() WHERE id = %s"
)

_LISTEN_QUERY = SQL("LISTEN {}")

_UNLISTEN_QUERY = SQL("UNLISTEN {}")



from __future__ import (
    annotations,
)

from datetime import (
    datetime,
)
from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DeclarativeModel,
)

from ..requests import (
    Request,
    ResponseException,
)


class ScheduledRequest(Request):
    """Scheduling Request class."""

    def __init__(self, scheduled_at: datetime):
        super().__init__()
        self._scheduled_at = scheduled_at

    @property
    def user(self) -> Optional[UUID]:
        """The user of the request.

        :return: Always return ``None`` as scheduled request are launched by the system.
        """
        return None

    async def content(self, **kwargs) -> ScheduledRequestContent:
        """Get the request content.

        :param kwargs: Additional named arguments.
        :return: A ``ScheduledRequestContent` intance.`.
        """
        return self._content

    def __eq__(self, other: Request) -> bool:
        return isinstance(other, type(self)) and self._content == other._content

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._content!r})"

    @property
    def _content(self) -> ScheduledRequestContent:
        return ScheduledRequestContent(self._scheduled_at)


class ScheduledRequestContent(DeclarativeModel):
    """Scheduling Request Content class."""

    scheduled_at: datetime


class ScheduledResponseException(ResponseException):
    """Scheduled Response Exception class."""



from __future__ import (
    annotations,
)

from .definitions import (
    BrokerCommandEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
    PeriodicEventEnrouteDecorator,
    RestCommandEnrouteDecorator,
    RestQueryEnrouteDecorator,
)


class BrokerEnroute:
    """Broker Enroute class"""

    command = BrokerCommandEnrouteDecorator
    query = BrokerQueryEnrouteDecorator
    event = BrokerEventEnrouteDecorator


class RestEnroute:
    """Rest Enroute class"""

    command = RestCommandEnrouteDecorator
    query = RestQueryEnrouteDecorator


class PeriodicEnroute:
    """Periodic Enroute class."""

    event = PeriodicEventEnrouteDecorator


class Enroute:
    """Enroute decorator main class"""

    broker = BrokerEnroute
    rest = RestEnroute
    periodic = PeriodicEnroute


enroute = Enroute



from asyncio import (
    gather,
)
from collections import (
    defaultdict,
)
from functools import (
    partial,
)
from inspect import (
    isawaitable,
)
from typing import (
    Awaitable,
    Callable,
    Optional,
    Type,
    Union,
)

from minos.common import (
    import_module,
)

from ..exceptions import (
    MinosRedefinedEnrouteDecoratorException,
)
from ..requests import (
    Request,
    Response,
)
from .analyzers import (
    EnrouteAnalyzer,
)
from .definitions import (
    BrokerEnrouteDecorator,
    EnrouteDecorator,
    EnrouteDecoratorKind,
    PeriodicEnrouteDecorator,
    RestEnrouteDecorator,
)

Handler = Callable[[Request], Awaitable[Optional[Response]]]


class EnrouteBuilder:
    """Enroute builder class."""

    def __init__(
        self, *classes: Union[str, Type], middleware: Optional[Union[str, Callable, list[Union[str, Callable]]]] = None
    ):
        if middleware is None:
            middleware = tuple()
        if not isinstance(middleware, (tuple, list)):
            middleware = [middleware]

        middleware = list(map(lambda fn: fn if not isinstance(fn, str) else import_module(fn), middleware))

        classes = tuple((class_ if not isinstance(class_, str) else import_module(class_)) for class_ in classes)

        self.classes = classes
        self.middleware = middleware

    def get_rest_command_query(self, **kwargs) -> dict[RestEnrouteDecorator, Handler]:
        """Get the rest handlers for commands and queries.

        :return: A dictionary with decorator classes as keys and callable handlers as values.
        """
        # noinspection PyTypeChecker
        return self._build("get_rest_command_query", **kwargs)

    def get_broker_command_query_event(self, **kwargs) -> dict[BrokerEnrouteDecorator, Handler]:
        """Get the broker handlers for commands, queries and events.

        :return: A dictionary with decorator classes as keys and callable handlers as values.
        """
        # noinspection PyTypeChecker
        return self._build("get_broker_command_query_event", **kwargs)

    def get_broker_command_query(self, **kwargs) -> dict[BrokerEnrouteDecorator, Handler]:
        """Get the broker handlers for commands and queries.

        :return: A dictionary with decorator classes as keys and callable handlers as values.
        """
        # noinspection PyTypeChecker
        return self._build("get_broker_command_query", **kwargs)

    def get_broker_event(self, **kwargs) -> dict[BrokerEnrouteDecorator, Handler]:
        """Get the broker handlers for events.

        :return: A dictionary with decorator classes as keys and callable handlers as values.
        """
        # noinspection PyTypeChecker
        return self._build("get_broker_event", **kwargs)

    def get_periodic_event(self, **kwargs) -> dict[PeriodicEnrouteDecorator, Handler]:
        """Get the periodic handlers for events.

        :return: A dictionary with decorator classes as keys and callable handlers as values.
        """
        # noinspection PyTypeChecker
        return self._build("get_periodic_event", **kwargs)

    def _build(self, method_name: str, **kwargs) -> dict[EnrouteDecorator, Handler]:
        def _flatten(decorator: EnrouteDecorator, fns: set[Handler]) -> Handler:
            if len(fns) == 1:
                return next(iter(fns))

            if decorator.KIND != EnrouteDecoratorKind.Event:
                raise MinosRedefinedEnrouteDecoratorException(f"{decorator!r} can be used only once.")

            async def _fn(*ag, **kw):
                return await gather(*(fn(*ag, **kw) for fn in fns))

            return _fn

        return {
            decorator: _flatten(decorator, fns)
            for decorator, fns in self._build_all_classes(method_name, **kwargs).items()
        }

    def _build_all_classes(self, method_name: str, **kwargs) -> dict[EnrouteDecorator, set[Handler]]:
        decomposed_handlers = defaultdict(set)
        for class_ in self.classes:
            self._build_one_class(class_, method_name, decomposed_handlers, **kwargs)
        return decomposed_handlers

    def _build_one_class(
        self, class_: type, method_name: str, ans: dict[EnrouteDecorator, set[Handler]], **kwargs
    ) -> None:
        analyzer = EnrouteAnalyzer(class_, **kwargs)
        mapping = getattr(analyzer, method_name)()

        for name, decorators in mapping.items():
            for decorator in decorators:
                ans[decorator].add(self._build_one_method(class_, name, decorator.pre_fn_name, decorator.post_fn_name))

    def _build_one_method(self, class_: type, name: str, pref_fn_name: str, post_fn_name: str, **kwargs) -> Handler:
        instance = class_(**kwargs)
        fn = getattr(instance, name)
        pre_fn = getattr(instance, pref_fn_name, None)
        post_fn = getattr(instance, post_fn_name, None)

        async def _wrapped_fn(request: Request) -> Optional[Response]:
            if pre_fn is not None:
                request = pre_fn(request)
                if isawaitable(request):
                    request = await request

            response = fn(request)
            if isawaitable(response):
                response = await response

            if post_fn is not None:
                response = post_fn(response)
                if isawaitable(response):
                    response = await response

            return response

        for middleware_fn in reversed(self.middleware):
            _wrapped_fn = partial(middleware_fn, inner=_wrapped_fn)

        return _wrapped_fn



from .abc import (
    EnrouteDecorator,
)
from .broker import (
    BrokerCommandEnrouteDecorator,
    BrokerEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
)
from .kinds import (
    EnrouteDecoratorKind,
)
from .periodic import (
    PeriodicEnrouteDecorator,
    PeriodicEventEnrouteDecorator,
)
from .rest import (
    RestCommandEnrouteDecorator,
    RestEnrouteDecorator,
    RestQueryEnrouteDecorator,
)



from inspect import (
    getmembers,
    isfunction,
    ismethod,
)
from typing import (
    Callable,
    Optional,
    Type,
    Union,
)

from minos.common import (
    MinosConfig,
    import_module,
)

from .definitions import (
    BrokerCommandEnrouteDecorator,
    BrokerEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
    EnrouteDecorator,
    PeriodicEventEnrouteDecorator,
    RestCommandEnrouteDecorator,
    RestEnrouteDecorator,
    RestQueryEnrouteDecorator,
)


class EnrouteAnalyzer:
    """Search decorators in specified class"""

    # noinspection PyUnusedLocal
    def __init__(self, decorated: Union[str, Type], config: Optional[MinosConfig] = None, **kwargs):
        if isinstance(decorated, str):
            decorated = import_module(decorated)

        self.decorated = decorated
        self.config = config

    def get_rest_command_query(self) -> dict[str, set[RestEnrouteDecorator]]:
        """Returns rest's command and query values.

        :return: A mapping with functions as keys and a sets of decorators as values.
        """
        # noinspection PyTypeChecker
        return self._get_items({RestCommandEnrouteDecorator, RestQueryEnrouteDecorator})

    def get_broker_command_query_event(self) -> dict[str, set[BrokerEnrouteDecorator]]:
        """Returns broker's command, query and event values.

        :return: A mapping with functions as keys and a sets of decorators as values.
        """
        # noinspection PyTypeChecker
        return self._get_items(
            {BrokerCommandEnrouteDecorator, BrokerQueryEnrouteDecorator, BrokerEventEnrouteDecorator}
        )

    def get_broker_command_query(self) -> dict[str, set[BrokerEnrouteDecorator]]:
        """Returns broker's command and query values.

        :return: A mapping with functions as keys and a sets of decorators as values.
        """
        # noinspection PyTypeChecker
        return self._get_items({BrokerCommandEnrouteDecorator, BrokerQueryEnrouteDecorator})

    def get_broker_event(self) -> dict[str, set[BrokerEnrouteDecorator]]:
        """Returns broker's event values.

        :return: A mapping with functions as keys and a sets of decorators as values.
        """
        # noinspection PyTypeChecker
        return self._get_items({BrokerEventEnrouteDecorator})

    def get_periodic_event(self) -> dict[str, set[PeriodicEventEnrouteDecorator]]:
        """Returns periodic event values.

        :return: A mapping with functions as keys and a sets of decorators as values.
        """
        # noinspection PyTypeChecker
        return self._get_items({PeriodicEventEnrouteDecorator})

    def _get_items(self, expected_types: set[Type[EnrouteDecorator]]) -> dict[str, set[EnrouteDecorator]]:
        items = dict()
        for fn, decorators in self.get_all().items():
            decorators = {decorator for decorator in decorators if type(decorator) in expected_types}
            if len(decorators):
                items[fn] = decorators
        return items

    def get_all(self) -> dict[str, set[EnrouteDecorator]]:
        """Get all functions decorated with enroute decorators.

        :return: A mapping with functions as keys and a sets of decorators as values.
        """
        fn: Callable = getattr(self.decorated, "__get_enroute__", self._get_all)
        return fn(config=self.config)

    # noinspection PyUnusedLocal
    def _get_all(self, *args, **kwargs) -> dict[str, set[EnrouteDecorator]]:
        result = dict()
        for name, fn in getmembers(self.decorated, predicate=lambda x: ismethod(x) or isfunction(x)):
            if not hasattr(fn, "__decorators__"):
                continue
            result[name] = fn.__decorators__
        return result



from abc import (
    ABC,
)
from typing import (
    Final,
    Iterable,
    Union,
)

from crontab import (
    CronTab,
)

from .abc import (
    EnrouteDecorator,
)
from .kinds import (
    EnrouteDecoratorKind,
)


class PeriodicEnrouteDecorator(EnrouteDecorator, ABC):
    """Periodic Enroute class"""

    def __init__(self, crontab: Union[str, CronTab]):
        if isinstance(crontab, str):
            crontab = CronTab(crontab)
        self.crontab = crontab

    def __iter__(self) -> Iterable:
        yield from (self.crontab,)

    def __hash__(self):
        return hash(tuple((s if not isinstance(s, CronTab) else s.matchers) for s in self))


class PeriodicEventEnrouteDecorator(PeriodicEnrouteDecorator):
    """Periodic Event Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Event



from .analyzers import (
    EnrouteAnalyzer,
)
from .api import (
    enroute,
)
from .builders import (
    EnrouteBuilder,
)
from .definitions import (
    BrokerCommandEnrouteDecorator,
    BrokerEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
    EnrouteDecorator,
    EnrouteDecoratorKind,
    PeriodicEnrouteDecorator,
    PeriodicEventEnrouteDecorator,
    RestCommandEnrouteDecorator,
    RestEnrouteDecorator,
    RestQueryEnrouteDecorator,
)



from __future__ import (
    annotations,
)

import logging
from contextvars import (
    Token,
)
from typing import (
    AsyncContextManager,
    Optional,
)
from uuid import (
    uuid4,
)

from dependency_injector.wiring import (
    Provide,
    inject,
)
from kafka import (
    KafkaAdminClient,
)
from kafka.admin import (
    NewTopic,
)

from minos.common import (
    MinosConfig,
    MinosPool,
    NotProvidedException,
)

from ..handlers import (
    BrokerConsumer,
)
from ..messages import (
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
)
from ..publishers import (
    BrokerPublisher,
)
from .brokers import (
    DynamicBroker,
)

logger = logging.getLogger(__name__)


class DynamicBrokerPool(MinosPool):
    """Dynamic Broker Pool class."""

    def __init__(
        self,
        config: MinosConfig,
        client: KafkaAdminClient,
        consumer: BrokerConsumer,
        publisher: BrokerPublisher,
        maxsize: int = 5,
        recycle: Optional[int] = 3600,
        *args,
        **kwargs,
    ):
        super().__init__(maxsize=maxsize, recycle=recycle, *args, **kwargs)
        self.config = config
        self.client = client
        self.consumer = consumer
        self.publisher = publisher

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> DynamicBrokerPool:
        kwargs["client"] = KafkaAdminClient(bootstrap_servers=f"{config.broker.host}:{config.broker.port}")
        kwargs["consumer"] = cls._get_consumer(**kwargs)
        kwargs["publisher"] = cls._get_publisher(**kwargs)
        return cls(config, **kwargs)

    async def _destroy(self) -> None:
        await super()._destroy()
        self.client.close()

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_consumer(
        consumer: Optional[BrokerConsumer] = None,
        broker_consumer: BrokerConsumer = Provide["broker_consumer"],
        **kwargs,
    ) -> BrokerConsumer:
        if consumer is None:
            consumer = broker_consumer
        if consumer is None or isinstance(consumer, Provide):
            raise NotProvidedException(f"A {BrokerConsumer!r} object must be provided.")
        return consumer

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_publisher(
        publisher: Optional[BrokerPublisher] = None,
        broker_publisher: BrokerPublisher = Provide["broker_publisher"],
        **kwargs,
    ) -> BrokerPublisher:
        if publisher is None:
            publisher = broker_publisher
        if publisher is None or isinstance(publisher, Provide):
            raise NotProvidedException(f"A {BrokerPublisher!r} object must be provided.")
        return publisher

    async def _create_instance(self) -> DynamicBroker:
        topic = str(uuid4()).replace("-", "")
        await self._create_reply_topic(topic)
        await self._subscribe_reply_topic(topic)
        instance = DynamicBroker.from_config(self.config, topic=topic, publisher=self.publisher)
        await instance.setup()
        return instance

    async def _destroy_instance(self, instance: DynamicBroker):
        await instance.destroy()
        await self._unsubscribe_reply_topic(instance.topic)
        await self._delete_reply_topic(instance.topic)

    async def _create_reply_topic(self, topic: str) -> None:
        logger.info(f"Creating {topic!r} topic...")
        self.client.create_topics([NewTopic(name=topic, num_partitions=1, replication_factor=1)])

    async def _delete_reply_topic(self, topic: str) -> None:
        logger.info(f"Deleting {topic!r} topic...")
        self.client.delete_topics([topic])

    async def _subscribe_reply_topic(self, topic: str) -> None:
        await self.consumer.add_topic(topic)

    async def _unsubscribe_reply_topic(self, topic: str) -> None:
        await self.consumer.remove_topic(topic)

    def acquire(self, *args, **kwargs) -> AsyncContextManager:
        """Acquire a new instance wrapped on an asynchronous context manager.

        :return: An asynchronous context manager.
        """
        return _ReplyTopicContextManager(super().acquire())


class _ReplyTopicContextManager:
    _token: Optional[Token]

    def __init__(self, wrapper: AsyncContextManager[DynamicBroker]):
        self.wrapper = wrapper
        self._token = None

    async def __aenter__(self) -> DynamicBroker:
        handler = await self.wrapper.__aenter__()
        self._token = REQUEST_REPLY_TOPIC_CONTEXT_VAR.set(handler.topic)
        return handler

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        REQUEST_REPLY_TOPIC_CONTEXT_VAR.reset(self._token)
        await self.wrapper.__aexit__(exc_type, exc_val, exc_tb)



from __future__ import (
    annotations,
)

from contextvars import (
    ContextVar,
)
from enum import (
    Enum,
    IntEnum,
)
from typing import (
    Any,
    Final,
    Optional,
)
from uuid import (
    UUID,
    uuid4,
)

from minos.common import (
    DeclarativeModel,
)

REQUEST_REPLY_TOPIC_CONTEXT_VAR: Final[ContextVar[Optional[str]]] = ContextVar("reply_topic", default=None)
REQUEST_HEADERS_CONTEXT_VAR: Final[ContextVar[Optional[dict[str, str]]]] = ContextVar("headers", default=None)


class BrokerMessage(DeclarativeModel):
    """Broker Message class."""

    topic: str
    data: Any
    identifier: UUID
    reply_topic: Optional[str]
    user: Optional[UUID]
    status: BrokerMessageStatus
    strategy: BrokerMessageStrategy
    headers: dict[str, str]

    def __init__(
        self,
        topic: str,
        data: Any,
        *,
        identifier: Optional[UUID] = None,
        status: Optional[BrokerMessageStatus] = None,
        strategy: Optional[BrokerMessageStrategy] = None,
        headers: Optional[dict[str, str]] = None,
        **kwargs
    ):
        if identifier is None:
            identifier = uuid4()
        if status is None:
            status = BrokerMessageStatus.SUCCESS
        if strategy is None:
            strategy = BrokerMessageStrategy.UNICAST
        if headers is None:
            headers = dict()
        super().__init__(
            topic=topic, data=data, identifier=identifier, status=status, strategy=strategy, headers=headers, **kwargs
        )

    @property
    def ok(self) -> bool:
        """Check if the reply is okay or not.

        :return: ``True`` if the reply is okay or ``False`` otherwise.
        """
        return self.status == BrokerMessageStatus.SUCCESS


class BrokerMessageStatus(IntEnum):
    """Broker Message Status class."""

    SUCCESS = 200
    ERROR = 400
    SYSTEM_ERROR = 500


class BrokerMessageStrategy(str, Enum):
    """Broker Message Strategy class"""

    UNICAST = "unicast"
    MULTICAST = "multicast"



from __future__ import (
    annotations,
)

import logging

from psycopg2.sql import (
    SQL,
)

from minos.common import (
    PostgreSqlMinosDatabase,
)

logger = logging.getLogger(__name__)


class BrokerPublisherSetup(PostgreSqlMinosDatabase):
    """Broker Publisher Setup class."""

    async def _setup(self) -> None:
        await self._create_broker_table()

    async def _create_broker_table(self) -> None:
        await self.submit_query(_CREATE_TABLE_QUERY, lock=hash("producer_queue"))


_CREATE_TABLE_QUERY = SQL(
    "CREATE TABLE IF NOT EXISTS producer_queue ("
    "id BIGSERIAL NOT NULL PRIMARY KEY, "
    "topic VARCHAR(255) NOT NULL, "
    "data BYTEA NOT NULL, "
    "strategy VARCHAR(255) NOT NULL, "
    "retry INTEGER NOT NULL DEFAULT 0, "
    "created_at TIMESTAMPTZ NOT NULL DEFAULT NOW(), "
    "updated_at TIMESTAMPTZ NOT NULL DEFAULT NOW())"
)



from .abc import (
    BrokerHandlerSetup,
)
from .consumers import (
    BrokerConsumer,
)
from .entries import (
    BrokerHandlerEntry,
)
from .handlers import (
    BrokerHandler,
)
from .requests import (
    BrokerRequest,
    BrokerResponse,
    BrokerResponseException,
)
from .services import (
    BrokerConsumerService,
    BrokerHandlerService,
)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from contextvars import (
    ContextVar,
)
from inspect import (
    isawaitable,
)
from typing import (
    Any,
    Callable,
    Final,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    AvroDataEncoder,
)

from .exceptions import (
    MinosException,
)

REQUEST_USER_CONTEXT_VAR: Final[ContextVar[Optional[UUID]]] = ContextVar("user", default=None)


class Request(ABC):
    """Request interface."""

    @property
    @abstractmethod
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        raise NotImplementedError

    @abstractmethod
    async def content(self, **kwargs) -> Any:
        """Get the request content.

        :param kwargs: Additional named arguments.
        :return: A list of instances.
        """
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, other: Request) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError


class WrappedRequest(Request):
    """Wrapped Request class."""

    def __init__(self, base: Request, action: Callable[[Any, ...], Any]):
        self.base = base
        self.action = action
        self._content = None

    @property
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        return self.base.user

    async def content(self, **kwargs) -> Any:
        """Get the request content.

        :param kwargs: Additional named arguments.
        :return: A list of instances.
        """
        if self._content is None:
            content = self.action(await self.base.content(), **kwargs)
            if isawaitable(content):
                content = await content
            self._content = content
        return self._content

    def __eq__(self, other: WrappedRequest) -> bool:
        return type(self) == type(other) and self.base == other.base and self.action == other.action

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.base!r}, {self.action!r})"


class Response:
    """Response definition."""

    __slots__ = "_data"

    def __init__(self, data: Any):
        self._data = data

    # noinspection PyUnusedLocal
    async def content(self, **kwargs) -> Any:
        """Response content.

        :param kwargs: Additional named arguments.
        :return: A list of items.
        """
        return self._data

    # noinspection PyUnusedLocal
    async def raw_content(self, **kwargs) -> Any:
        """Raw response content.

        :param kwargs: Additional named arguments.
        :return: A list of raw items.
        """
        return AvroDataEncoder(self._data).build()

    def __eq__(self, other: Response) -> bool:
        return type(self) == type(other) and self._data == other._data

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._data!r})"

    def __hash__(self):
        return hash(self._data)


class ResponseException(MinosException):
    """Response Exception class."""



from __future__ import (
    annotations,
)

from collections import (
    defaultdict,
)
from json import (
    JSONDecodeError,
)
from typing import (
    Any,
    Optional,
    Type,
    Union,
)
from uuid import (
    UUID,
)

from aiohttp import (
    web,
)
from cached_property import (
    cached_property,
)

from minos.common import (
    MinosImportException,
    Model,
    ModelType,
    TypeHintBuilder,
    import_module,
)

from ..requests import (
    Request,
    Response,
    ResponseException,
)


class RestRequest(Request):
    """Rest Request class."""

    def __init__(self, request: web.Request):
        self.raw_request = request

    def __eq__(self, other: RestRequest) -> bool:
        return type(self) == type(other) and self.raw_request == other.raw_request

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.raw_request!r})"

    @cached_property
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        if "User" not in self.headers:
            return None
        return UUID(self.headers["User"])

    @property
    def headers(self) -> dict[str, str]:
        """Get the headers of the request.

        :return: A dictionary in which keys are ``str`` instances and values are ``str`` instances.
        """
        # noinspection PyTypeChecker
        return self.raw_request.headers

    async def content(self, model_type: Union[ModelType, Type[Model], str] = "Content", **kwargs) -> Any:
        """Get the request content.

        :param model_type: Optional ``ModelType`` that defines the request content structure.
        :param kwargs: Additional named arguments.
        :return: The command content.
        """
        data = await self._raw_json()
        data = [(entry | self.url_args | self.path_args) for entry in data]
        data = self._build_models(data, model_type)

        if len(data) == 1:
            return data[0]
        return data

    async def _raw_json(self) -> list[dict[str, Any]]:
        try:
            data = await self.raw_request.json()
        except JSONDecodeError:
            data = dict()

        if not isinstance(data, list):
            data = [data]
        return data

    @cached_property
    def url_args(self) -> dict[str, Any]:
        """Get the url arguments as a dictionary.

        :return: A dictionary instance.
        """
        args = defaultdict(list)
        for k, v in self._raw_url_args:
            args[k].append(v)
        args = {k: v if len(v) > 1 else v[0] for k, v in args.items()}
        return args

    @property
    def _raw_url_args(self):
        return self.raw_request.rel_url.query.items()  # pragma: no cover

    @cached_property
    def path_args(self) -> dict[str, Any]:
        """Get the path arguments as a dictionary.

        :return: A dictionary instance.
        """
        args = defaultdict(list)
        for k, v in self._raw_path_args:
            args[k].append(v)
        args = {k: v if len(v) > 1 else v[0] for k, v in args.items()}
        return args

    @property
    def _raw_path_args(self):
        return self.raw_request.match_info.items()  # pragma: no cover

    def _build_models(self, data: list[dict[str, Any]], model_type: Union[ModelType, Type[Model], str]) -> list[Model]:
        return [self._build_one_model(entry, model_type) for entry in data]

    @staticmethod
    def _build_one_model(entry: dict[str, Any], model_type: Union[ModelType, Type[Model], str]) -> Model:
        if isinstance(model_type, str):
            try:
                model_type = import_module(model_type)
            except MinosImportException:
                type_hints = {k: TypeHintBuilder(v).build() for k, v in entry.items()}
                model_type = ModelType.build(model_type, type_hints)

        return model_type(**entry)


class RestResponse(Response):
    """Rest Response class."""


class RestResponseException(ResponseException):
    """Rest Response Exception class."""



from __future__ import (
    annotations,
)

import logging

from aiomisc import (
    Service,
)
from cached_property import (
    cached_property,
)

from .producers import (
    BrokerProducer,
)

logger = logging.getLogger(__name__)


class BrokerProducerService(Service):
    """Broker Producer Service class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_kwargs = kwargs

    async def start(self) -> None:
        """Start the service execution.

        :return: This method does not return anything.
        """
        await self.dispatcher.setup()

        try:
            self.start_event.set()
        except RuntimeError:
            logger.warning("Runtime is not properly setup.")

        await self.dispatcher.dispatch_forever()

    async def stop(self, err: Exception = None) -> None:
        """Stop the service execution.

        :param err: Optional exception that stopped the execution.
        :return: This method does not return anything.
        """
        await self.dispatcher.destroy()

    @cached_property
    def dispatcher(self) -> BrokerProducer:
        """Get the service dispatcher.

        :return: A ``Producer`` instance.
        """
        return BrokerProducer.from_config(**self._init_kwargs)



from __future__ import (
    annotations,
)

import logging
from contextlib import (
    suppress,
)
from itertools import (
    chain,
)
from typing import (
    Any,
    NoReturn,
    Optional,
)

from aiokafka import (
    AIOKafkaConsumer,
)
from kafka.errors import (
    IllegalStateError,
    KafkaError,
)
from psycopg2.sql import (
    SQL,
    Identifier,
)

from minos.common import (
    BROKER,
    MinosConfig,
)

from ...decorators import (
    EnrouteAnalyzer,
)
from .abc import (
    BrokerHandlerSetup,
)

logger = logging.getLogger(__name__)


class BrokerConsumer(BrokerHandlerSetup):
    """Broker Consumer class."""

    __slots__ = "_topics", "_broker", "_client"

    def __init__(
        self,
        topics: set[str] = None,
        broker: Optional[BROKER] = None,
        client: Optional[AIOKafkaConsumer] = None,
        group_id: Optional[str] = "default",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if topics is None:
            topics = set()
        self._topics = set(topics)
        self._broker = broker
        self._client = client
        self._group_id = group_id

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> BrokerConsumer:
        topics = set()

        for name in config.services:
            decorators = EnrouteAnalyzer(name, config).get_broker_command_query_event()
            topics |= {decorator.topic for decorator in chain(*decorators.values())}

        # noinspection PyProtectedMember
        return cls(
            topics=topics, broker=config.broker, group_id=config.service.name, **config.broker.queue._asdict(), **kwargs
        )

    async def _setup(self) -> None:
        await super()._setup()
        await self.client.start()

    async def _destroy(self) -> None:
        try:
            await self.client.stop()
        except KafkaError:  # pragma: no cover
            pass
        await super()._destroy()

    @property
    def topics(self) -> set[str]:
        """Topics getter.

        :return: A list of string values.
        """
        return self._topics

    async def add_topic(self, topic: str) -> None:
        """Add a topic to the consumer's subscribed topics.

        :param topic: Name of the topic to be added.
        :return: This method does not return anything.
        """
        self._topics.add(topic)
        self.client.subscribe(topics=list(self._topics))

    async def remove_topic(self, topic: str) -> None:
        """Remove a topic from the consumer's subscribed topics.

        :param topic: Name of the topic to be removed.
        :return: This method does not return anything.
        """
        self._topics.remove(topic)
        if len(self._topics):
            self.client.subscribe(topics=list(self._topics))
        else:
            self.client.unsubscribe()

    @property
    def client(self) -> AIOKafkaConsumer:
        """Get the kafka consumer client.

        :return: An ``AIOKafkaConsumer`` instance.
        """
        if self._client is None:  # pragma: no cover
            self._client = AIOKafkaConsumer(
                *self._topics,
                bootstrap_servers=f"{self._broker.host}:{self._broker.port}",
                group_id=self._group_id,
                enable_auto_commit=False,
                auto_offset_reset="earliest",
            )
        return self._client

    async def dispatch(self) -> NoReturn:
        """Perform a dispatching step.

        :return: This method does not return anything.
        """
        await self.handle_message(self.client)

    async def handle_message(self, consumer: Any) -> None:
        """Message consumer.

        It consumes the messages and sends them for processing.

        Args:
            consumer: Kafka Consumer instance (at the moment only Kafka consumer is supported).
        """

        async for message in consumer:
            await self.handle_single_message(message)
            with suppress(IllegalStateError):
                await consumer.commit()

    async def handle_single_message(self, message):
        """Handle Kafka messages.

        Add the message to the event_queue table.

        Args:
            message: Kafka message.

        Raises:
            Exception: An error occurred inserting record.
        """
        logger.debug(f"Consuming message with {message.topic!s} topic...")

        return await self.enqueue(message.topic, message.partition, message.value)

    async def enqueue(self, topic: str, partition: int, binary: bytes) -> int:
        """Insert row into queue table.

        Retrieves number of affected rows and row ID.

        Args:
            topic: Kafka topic. Example: "TicketAdded"
            partition: Kafka partition number.
            binary: Broker Message in bytes.

        Returns:
            Queue ID.

            Example: 12

        Raises:
            Exception: An error occurred inserting record.
        """
        row = await self.submit_query_and_fetchone(_INSERT_QUERY, (topic, partition, binary))
        await self.submit_query(_NOTIFY_QUERY.format(Identifier(topic)))

        return row[0]


_INSERT_QUERY = SQL("INSERT INTO consumer_queue (topic, partition, data) VALUES (%s, %s, %s) RETURNING id")

_NOTIFY_QUERY = SQL("NOTIFY {}")



import logging

from .abc import (
    DiscoveryClient,
)

logger = logging.getLogger(__name__)


class MinosDiscoveryClient(DiscoveryClient):
    """Minos Discovery Client class."""

    async def subscribe(
        self,
        host: str,
        port: int,
        name: str,
        endpoints: list[dict[str, str]],
        retry_tries: int = 3,
        retry_delay: float = 5,
    ) -> None:
        """Perform a subscription query.

        :param host: The ip of the microservice to be subscribed.
        :param port: The port of the microservice to be subscribed.
        :param name: The name of the microservice to be subscribed.
        :param endpoints: List of endpoints exposed by the microservice.
        :param retry_tries: Number of attempts before raising a failure exception.
        :param retry_delay: Seconds to wait between attempts.
        :return: This method does not return anything.
        """
        endpoint = f"{self.route}/microservices/{name}"
        service_metadata = {
            "address": host,
            "port": port,
            "endpoints": [[endpoint["method"], endpoint["url"]] for endpoint in endpoints],
        }

        await self._rest_subscribe(endpoint, service_metadata, host, port, name, endpoints, retry_tries, retry_delay)

    async def unsubscribe(self, name: str, retry_tries: int = 3, retry_delay: float = 5) -> None:
        """Perform an unsubscribe query.

        :param name: The name of the microservice to be unsubscribed.
        :param retry_tries: Number of attempts before raising a failure exception.
        :param retry_delay: Seconds to wait between attempts.
        :return: This method does not return anything.
        """
        endpoint = f"{self.route}/microservices/{name}"

        await self._rest_unsubscribe(endpoint, name, retry_tries, retry_delay)



from .requests import (
    ScheduledRequest,
    ScheduledRequestContent,
    ScheduledResponseException,
)
from .schedulers import (
    PeriodicTask,
    PeriodicTaskScheduler,
)
from .services import (
    PeriodicTaskSchedulerService,
)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from inspect import (
    iscoroutinefunction,
)
from typing import (
    Awaitable,
    Callable,
    Final,
    Iterable,
    Optional,
    Union,
)

from ...exceptions import (
    MinosMultipleEnrouteDecoratorKindsException,
)
from ...requests import (
    Request,
    Response,
)
from .kinds import (
    EnrouteDecoratorKind,
)

Adapter = Callable[[Request], Union[Optional[Response], Awaitable[Optional[Response]]]]


class EnrouteDecorator(ABC):
    """Base Decorator class."""

    # noinspection PyFinal
    KIND: Final[EnrouteDecoratorKind]

    def __call__(self, fn: Adapter) -> Adapter:
        if iscoroutinefunction(fn):

            async def _wrapper(*args, **kwargs) -> Optional[Response]:
                return await fn(*args, **kwargs)

        else:

            def _wrapper(*args, **kwargs) -> Optional[Response]:
                return fn(*args, **kwargs)

        _wrapper.__decorators__ = getattr(fn, "__decorators__", set())
        _wrapper.__decorators__.add(self)
        kinds = set(decorator.KIND for decorator in _wrapper.__decorators__)
        if len(kinds) > 1:
            raise MinosMultipleEnrouteDecoratorKindsException(
                f"There are multiple kinds but only one is allowed: {kinds}"
            )
        _wrapper.__base_func__ = getattr(fn, "__base_func__", fn)

        return _wrapper

    def __repr__(self):
        args = ", ".join(map(repr, self))
        return f"{type(self).__name__}({args})"

    def __eq__(self, other: EnrouteDecorator) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    @abstractmethod
    def __iter__(self) -> Iterable:
        raise NotImplementedError

    @property
    def pre_fn_name(self) -> str:
        """Get the pre execution function name.

        :return: A string value containing the function name.
        """
        return self.KIND.pre_fn_name

    @property
    def post_fn_name(self) -> str:
        """Get the post execution function name.

        :return: A string value containing the function name.
        """
        return self.KIND.post_fn_name



from aiohttp import (
    web,
)
from aiomisc.service.aiohttp import (
    AIOHTTPService,
)

from .handlers import (
    RestHandler,
)


class RestService(AIOHTTPService):
    """
    Rest Interface

    Expose REST Interface handler using aiomisc AIOHTTPService.

    """

    def __init__(self, **kwargs):
        self.handler = RestHandler.from_config(**kwargs)
        super().__init__(**(kwargs | {"address": self.handler.host, "port": self.handler.port}))

    async def create_application(self) -> web.Application:
        """Create the web application.

        :return: A ``web.Application`` instance.
        """
        return self.handler.get_app()  # pragma: no cover
